# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtCore

from prettyqt import core


QtCore.QAbstractListModel.__bases__ = (core.AbstractItemModel,)


class AbstractListModel(QtCore.QAbstractListModel):
    pass



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtCore
from prettyqt import core


class RegularExpressionMatchIterator(QtCore.QRegularExpressionMatchIterator):

    def __repr__(self):
        return "RegularExpressionMatchIterator()"

    def __iter__(self):
        return self

    def __next__(self):
        if self.hasNext():
            return core.RegularExpressionMatch(self.next())
        raise StopIteration

    def peek_next(self):
        return core.RegularExpressionMatch(self.peekNext())


if __name__ == "__main__":
    reg = RegularExpressionMatchIterator()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `prettyqt` package."""

from prettyqt import syntaxhighlighters
import re


def test_jsonhighlighter():
    highlighter = syntaxhighlighters.JsonHighlighter()
    highlighter.highlightBlock('{"a": "b"}')


def test_pythonhighlighter():
    highlighter = syntaxhighlighters.PythonHighlighter()
    highlighter.highlightBlock("def test(): pass")


def test_yamlhighlighter():
    highlighter = syntaxhighlighters.YamlHighlighter()
    highlighter.highlightBlock("---\ntest:\n  - hallo")


def test_xmlhighlighter():
    highlighter = syntaxhighlighters.XmlHighlighter()
    highlighter.highlightBlock("<xml>test</xml>")


def test_markdownhighlighter():
    highlighter = syntaxhighlighters.MarkdownHighlighter()
    highlighter.highlightBlock("### Headline")


def test_regexmatchhighlighter():
    highlighter = syntaxhighlighters.RegexMatchHighlighter()
    text = "a123"
    prog = re.compile("[0-9]")
    spans = [m.span() for m in prog.finditer(text)]
    highlighter.set_spans(spans)
    highlighter.highlightBlock(text)



# -*- coding: utf-8 -*-
"""
"""

import sys
import logging
import re
from abc import ABC, abstractmethod

from prettyqt import gui, widgets, constants
from prettyqt.utils import signallogger

# from SublimeText Regex:
# (?x:
#   (?:.? [<>=^])?     # fill align
#   [ +-]?             # sign
#   \#?                # alternate form
#   # technically, octal and hexadecimal integers are also supported as 'width'
#   \d*                # width
#   ,?                 # thousands separator
#   (?:\.\d+)?         # precision
#   [bcdeEfFgGnosxX%]? # type
# )


class Highlighter(ABC):
    placeholder: str
    color = "black"
    italic = False
    bold = False
    value = None

    def __init__(self, formatter):
        self.formatter = formatter
        self.format = gui.TextCharFormat(self.color, self.bold, self.italic)
        text = re.escape(self.placeholder)
        pat = fr"{text[:-1]}([ +-]?\#?\#?\d*,?(?:\.\d+)?[bcdeEfFgGnosxX%]?)"
        self.pattern = re.compile(pat)
        self.is_included = self.pattern.search(self.formatter._fmt) is not None

    def get_format(self, value) -> gui.TextCharFormat:
        return self.format

    @abstractmethod
    def format_string(self):
        raise NotImplementedError()


class AscTime(Highlighter):
    placeholder = "%(asctime)s"
    italic = True

    def format_string(self, record: logging.LogRecord) -> str:
        return self.formatter.formatTime(record)


class Message(Highlighter):
    placeholder = "%(message)s"
    bold = True

    def format_string(self, record: logging.LogRecord) -> str:
        if isinstance(record.msg, Exception):
            val = self.formatter.formatException(record.exc_info)
        else:
            val = record.msg % record.args
        if "\n" in val:
            val = f"\n{val}"
        return val


class FileName(Highlighter):
    placeholder = "%(filename)s"
    bold = True

    def format_string(self, record: logging.LogRecord) -> str:
        return record.filename


class FuncName(Highlighter):
    placeholder = "%(funcName)s"
    bold = True

    def format_string(self, record: logging.LogRecord) -> str:
        return record.funcName


class Module(Highlighter):
    placeholder = "%(module)s"
    bold = True

    def format_string(self, record: logging.LogRecord) -> str:
        return record.module


class Created(Highlighter):
    placeholder = "%(created)f"
    bold = True

    def format_string(self, record: logging.LogRecord) -> str:
        return str(record.created)


class LineNo(Highlighter):
    placeholder = "%(lineno)d"
    bold = True

    def format_string(self, record: logging.LogRecord) -> str:
        return str(record.lineno)


class Msecs(Highlighter):
    placeholder = "%(msecs)d"
    bold = True

    def format_string(self, record: logging.LogRecord) -> str:
        return str(record.msecs)


class Process(Highlighter):
    placeholder = "%(process)d"
    bold = True

    def format_string(self, record: logging.LogRecord) -> str:
        return str(record.process)


class Thread(Highlighter):
    placeholder = "%(thread)d"
    bold = True

    def format_string(self, record: logging.LogRecord) -> str:
        return str(record.thread)


class ThreadName(Highlighter):
    placeholder = "%(threadName)s"
    bold = True

    def format_string(self, record: logging.LogRecord) -> str:
        return record.threadName


class ProcessName(Highlighter):
    placeholder = "%(processName)s"
    bold = True

    def format_string(self, record: logging.LogRecord) -> str:
        return record.processName


class RelativeCreated(Highlighter):
    placeholder = "%(relativeCreated)s"
    bold = True

    def format_string(self, record: logging.LogRecord) -> str:
        return record.relativeCreated


class Name(Highlighter):
    placeholder = "%(name)s"
    bold = True

    def format_string(self, record: logging.LogRecord) -> str:
        return record.name


class PathName(Highlighter):
    placeholder = "%(pathname)s"
    bold = True

    def format_string(self, record: logging.LogRecord) -> str:
        return record.pathname


class LevelName(Highlighter):
    placeholder = "%(levelname)s"
    color = "red"
    formats = dict(DEBUG=gui.TextCharFormat(text_color="green", bold=True),
                   INFO=gui.TextCharFormat(text_color="blue", bold=True),
                   WARNING=gui.TextCharFormat(text_color="orange", bold=True),
                   CRITICAL=gui.TextCharFormat(text_color="darkorange", bold=True),
                   ERROR=gui.TextCharFormat(text_color="red", bold=True))

    def format_string(self, record: logging.LogRecord) -> str:
        return record.levelname

    def get_format(self, value) -> gui.TextCharFormat:
        return self.formats[value]


class LogTextEdit(widgets.PlainTextEdit):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_font("Consolas")
        self.formatter = None
        self.append_text(f"Python version: {sys.version}")
        logger = logging.getLogger()
        # self.handler = signallogger.LineSignalLogger()
        # self.handler.log_line.connect(self.append_text)
        self.handler = signallogger.RecordSignalLogger()
        self.handler.log_record.connect(self.append_record)
        self.handler.setLevel(logging.INFO)
        logger.addHandler(self.handler)
        fmt = logging.Formatter('%(asctime)s  %(levelname)i  %(message)s')
        self.set_formatter(fmt)

    def wheelEvent(self, event):
        """
        handle wheel event for zooming
        """
        if event.modifiers() & constants.CTRL_MOD:
            self.zoomIn() if event.angleDelta().y() > 0 else self.zoomOut()
        else:
            super().wheelEvent(event)

    def set_formatter(self, formatter):
        self.formatter = formatter
        self.handler.setFormatter(self.formatter)
        rules = [klass(self.formatter) for klass in Highlighter.__subclasses__()]
        self.rules = [r for r in rules if r.is_included]

    def append_record(self, record):
        start_of_line = len(self.text())
        self.append_text(self.formatter._fmt)
        old_fmt = self.textCursor().charFormat()
        with self.create_cursor() as c:
            c.move_position("end")
            c.move_position("start_of_block")
            start_pos = c.position()
            for r in self.rules:
                line_text = c.select_text(start_pos, "end_of_block")
                matches = list(r.pattern.finditer(line_text))
                for m in reversed(matches):
                    pos = m.start(0) + start_of_line
                    if start_of_line != 0:
                        pos += 1
                    end = pos + m.end(0) - m.start(0)
                    fmt_string = r.format_string(record)
                    try:
                        value = f"%{m.group(1)}" % fmt_string
                    except (TypeError, ValueError):
                        value = fmt_string
                    c.replace_text(pos, end, value)
                    fmt = r.get_format(fmt_string)
                    c.setCharFormat(fmt)
                    c.clearSelection()
                    c.setCharFormat(old_fmt)


if __name__ == "__main__":
    app = widgets.app()
    w = widgets.Widget()
    w.set_layout("vertical")
    widget = LogTextEdit()
    logger = logging.getLogger()

    def raise_exc():
        try:
            raise Exception("test")
        except Exception as e:
            logger.exception(e)

    w.box.add(widgets.PushButton("Raise", callback=raise_exc))
    w.box.add(widgets.PushButton("Debug", callback=lambda: logger.debug("Debug")))
    w.box.add(widgets.PushButton("Info", callback=lambda: logger.info("Info")))
    w.box.add(widgets.PushButton("Warning", callback=lambda: logger.warning("Warning")))
    w.box.add(widgets.PushButton("Critical",
                                 callback=lambda: logger.critical("Critical")))
    w.box.add(widget)
    w.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtWidgets, QtCore

from prettyqt import widgets


QtWidgets.QTableWidget.__bases__ = (widgets.TableView,)


class TableWidget(QtWidgets.QTableWidget):

    def sort(self, column=0, reverse=False):
        order = QtCore.Qt.DescendingOrder if reverse else QtCore.Qt.AscendingOrder
        self.sortItems(column, order)


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    widget = TableWidget()
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""

import docutils.core
from qtpy import QtWidgets
from typing import Union
import pathlib

from prettyqt import gui, widgets, core

QtWidgets.QTextBrowser.__bases__ = (widgets.TextEdit,)


class TextBrowser(QtWidgets.QTextBrowser):

    value_changed = core.Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setOpenExternalLinks(True)

    def __getstate__(self):
        return dict(text=self.text(),
                    enabled=self.isEnabled(),
                    font=gui.Font(self.font()))

    def __setstate__(self, state):
        self.__init__()
        self.setPlainText(state["text"])
        self.setEnabled(state.get("enabled", True))
        self.setFont(state["font"])

    # def dragEnterEvent(self, event):
    #     u = event.mimeData().urls()
    #     for url in u:
    #         file_path = os.path.abspath(url.toLocalFile())

    #         ext = file_path.split(".")[-1]
    #         if ext in ["txt", "md", "markdown"]:
    #             event.accept()
    #         else:
    #             event.ignore()

    # def dropEvent(self, event):
    #     event.accept()
    #     self.show_markdown_file(self.filePath)

    def set_markdown_file(self, file_path: Union[str, pathlib.Path]):
        if isinstance(file_path, str):
            file_path = pathlib.Path(file_path)
        with file_path.open() as f:
            file_content = f.read()
        self.set_markdown(file_content)

    def set_markdown(self, source: str):
        self.setMarkdown(source)

    def set_rst_file(self, file_path: Union[str, pathlib.Path]):
        if isinstance(file_path, str):
            file_path = pathlib.Path(file_path)
        with file_path.open() as f:
            file_content = f.read()
        self.set_rst(file_content)

    def set_rst(self, source: str):
        html = docutils.core.publish_string(source, writer_name="html")
        self.setHtml(str(html))


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    reader = TextBrowser()
    reader.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""

import bidict as bdct


class bidict(bdct.bidict):

    def __init__(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], dict):
            super().__init__(args[0])
        else:
            super().__init__(kwargs)



# -*- coding: utf-8 -*-
"""
"""

import regex as re
from prettyqt import core, widgets, custom_widgets, custom_validators


class RegexInput(widgets.Widget):

    value_changed = core.Signal()

    def __init__(self, show_flags=True, show_error=True, parent=None):
        super().__init__(parent=parent)
        self.set_layout("grid")
        self.lineedit = custom_widgets.SingleLineTextEdit()
        self.tb_flags = custom_widgets.BoolDictToolButton("Flags",
                                                          icon="mdi.flag-variant-outline")
        self.label_error = widgets.Label()
        self.label_error.setStyleSheet("color: #FF0000;")
        self.box[0, 0:1] = self.lineedit
        if show_flags:
            self.box[0, 2] = self.tb_flags
        if show_error:
            self.box[1, 0:2] = self.label_error
        self.tb_flags.triggered.connect(self._on_value_change)
        self.lineedit.textChanged.connect(self._on_value_change)
        val = custom_validators.RegexPatternValidator()
        self.lineedit.set_validator(val)
        dct = {"multiline": "MultiLine",
               "ignorecase": "Ignore case",
               "ascii": "ASCII-only matching",
               "dotall": "Dot matches newline",
               "verbose": "Ignore whitespace"}
        self._mapping = {"ignorecase": re.IGNORECASE,
                         "multiline": re.MULTILINE,
                         "ascii": re.ASCII,
                         "dotall": re.DOTALL,
                         "verbose": re.VERBOSE}
        self.tb_flags.set_dict(dct)

    def _on_value_change(self):
        self.value_changed.emit()
        if self.lineedit.is_valid():
            self.label_error.set_text("")
        else:
            message = self.lineedit.validator.error_message
            self.label_error.set_text(message)

    @property
    def pattern(self):
        return self.lineedit.text()

    @pattern.setter
    def pattern(self, value):
        self.lineedit.set_text(value)

    @property
    def compile_flags(self):
        ret_val = 0
        for identifier, flag in self._mapping.items():
            if self.tb_flags[identifier]:
                ret_val |= flag
        return ret_val

    @compile_flags.setter
    def compile_flags(self, value):
        for identifier, flag in self._mapping.items():
            self.tb_flags[identifier] = bool(value & flag)

    def set_value(self, value):
        if value is None:
            self.pattern = ""
            self.flags = 0
        if isinstance(value, str):
            self.pattern = value
            self.flags = 0
        else:
            self.pattern = value.pattern
            self.compile_flags = value.flags

    def get_value(self):
        return re.compile(self.pattern, self.compile_flags)

    def is_valid(self):
        return self.lineedit.is_valid()


if __name__ == "__main__":
    app = widgets.app()
    widget = RegexInput(show_flags=False, show_error=False)
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtWidgets

from prettyqt import widgets
from prettyqt.utils import bidict

SHADOWS = bidict(plain=QtWidgets.QFrame.Plain,
                 raised=QtWidgets.QFrame.Raised,
                 sunken=QtWidgets.QFrame.Sunken)

SHAPES = bidict(no_frame=QtWidgets.QFrame.NoFrame,
                box=QtWidgets.QFrame.Box,
                panel=QtWidgets.QFrame.Panel,
                styled_panel=QtWidgets.QFrame.StyledPanel,
                h_line=QtWidgets.QFrame.HLine,
                v_line=QtWidgets.QFrame.VLine,
                win_panel=QtWidgets.QFrame.WinPanel)

QtWidgets.QFrame.__bases__ = (widgets.Widget,)


class Frame(QtWidgets.QFrame):

    def set_frame_style(self, style: str):
        """set frame style

        Allowed values are "plain", "raised", "sunken"

        Args:
            style: frame style to use

        Raises:
            ValueError: style does not exist
        """
        if style not in SHADOWS:
            raise ValueError("invalid frame style")
        self.setFrameStyle(SHADOWS[style])

    def get_frame_style(self) -> str:
        """returns current frame style

        Possible values: "plain", "raised", "sunken"

        Returns:
            frame style
        """
        return SHADOWS.inv[self.frameStyle()]



# -*- coding: utf-8 -*-
"""
"""


from prettyqt import constants, widgets, core


class RegexMatchesModel(core.AbstractTableModel):

    HEADER = ["Start", "End", "Value", "Groups"]

    def __init__(self, matches, parent=None):
        super().__init__(parent=parent)
        self.matches = matches

    def columnCount(self, parent=None):
        return len(self.HEADER)

    def headerData(self, offset: int, orientation, role):
        if role == constants.DISPLAY_ROLE:
            if orientation == constants.HORIZONTAL:
                return self.HEADER[offset]

    def data(self, index, role):
        if not index.isValid():
            return None
        item = self.matches[index.row()]
        if role in [constants.DISPLAY_ROLE]:
            if index.column() == 0:
                return str(item.span()[0])
            if index.column() == 1:
                return str(item.span()[1])
            elif index.column() == 2:
                return repr(item.group())
            elif index.column() == 3:
                return str(len(item.groups()))
        if role in [constants.USER_ROLE]:
            return item.span()

    def rowCount(self, parent=None):
        """
        required override for AbstractitemModels
        """
        return len(self.matches)


if __name__ == "__main__":
    import re
    app = widgets.Application([])
    view = widgets.TableView()
    teststring = "fjdjkslfj355fjdkj 55 55454"
    regex = r"(?:[0-9])*.([0-9])"
    teststring = "SH_6208069141055_BC000388_20110412101855"
    regex = r"(?:([a-z]{2,})_)?(\d+)_([a-z]{2,}\d+)_(\d+)$"
    compiled = re.compile(regex, re.IGNORECASE)
    matches = list(compiled.finditer(teststring))
    model = RegexMatchesModel(matches)

    view.setModel(model)
    dlg = widgets.Dialog(layout="horizontal")
    dlg.add_widget(view)
    dlg.show_blocking()
    view.resize(500, 300)
    print(view.model())
    print(view.model().rowCount())
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtWidgets

from prettyqt import gui, widgets
from prettyqt.utils import colors

OPTIONS = dict(show_alpha=QtWidgets.QColorDialog.ShowAlphaChannel,
               no_buttons=QtWidgets.QColorDialog.NoButtons,
               no_native=QtWidgets.QColorDialog.DontUseNativeDialog)


QtWidgets.QColorDialog.__bases__ = (widgets.BaseDialog,)


class ColorDialog(QtWidgets.QColorDialog):

    def __getstate__(self):
        return dict(color=self.current_color())

    def __setstate__(self, state):
        self.__init__()
        if state["color"]:
            self.setCurrentColor(state["color"])

    @classmethod
    def get_color(cls,
                  preset: colors.ColorType = None,
                  allow_alpha: bool = False,
                  parent=None) -> gui.Color:
        preset = colors.get_color(preset)
        kwargs = dict(options=cls.ShowAlphaChannel) if allow_alpha else dict()
        return gui.Color(cls.getColor(preset, parent, **kwargs))

    def current_color(self) -> gui.Color:
        return gui.Color(self.currentColor())


if __name__ == "__main__":
    app = widgets.app()
    ColorDialog.get_color()
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtCore

from prettyqt import core

QtCore.QItemSelectionModel.__bases__ = (core.Object,)


class ItemSelectionModel(QtCore.QItemSelectionModel):
    pass



# -*- coding: utf-8 -*-
"""
"""

import datetime
from typing import Optional

from qtpy import QtWidgets

from prettyqt import widgets
from prettyqt.utils import bidict

SELECTION_MODES = bidict(none=QtWidgets.QCalendarWidget.NoSelection,
                         single=QtWidgets.QCalendarWidget.SingleSelection)

HEADER_FORMATS = bidict(none=QtWidgets.QCalendarWidget.NoVerticalHeader,
                        week_numbers=QtWidgets.QCalendarWidget.ISOWeekNumbers)


QtWidgets.QCalendarWidget.__bases__ = (widgets.Widget,)


class CalendarWidget(QtWidgets.QCalendarWidget):

    def __getstate__(self):
        return dict(date=self.selectedDate())

    def __setstate__(self, state):
        self.__init__()
        self.setSelectedDate(state["date"])

    def get_date(self) -> datetime.date:
        try:
            return self.selectedDate().toPyDate()  # pyqt5
        except (TypeError, AttributeError):
            return self.selectedDate().toPython()

    def get_value(self) -> datetime.date:
        return self.get_date()

    def set_value(self, value):
        self.setSelectedDate(value)

    def set_selection_mode(self, mode: Optional[str]):
        """set selection mode for given calendar widget

        Allowed values are "single" or "none"

        Args:
            mode: selection mode to use

        Raises:
            ValueError: mode does not exist
        """
        if mode is None:
            mode = "none"
        if mode not in SELECTION_MODES:
            raise ValueError("Format must be either 'single' or 'None'")
        self.setSelectionMode(SELECTION_MODES[mode])

    def get_selection_mode(self) -> str:
        """returns current selection mode

        Possible values: "single" or "none"

        Returns:
            selection mode
        """
        return SELECTION_MODES.inv[self.selectionMode()]


if __name__ == "__main__":
    app = widgets.app()
    w = CalendarWidget()
    w.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""

from qtpy.QtCharts import QtCharts
from prettyqt import charts

QtCharts.QXYSeries.__bases__ = (charts.AbstractSeries,)


class XYSeries(QtCharts.QXYSeries):

    """
    QXYSeries with some custom properties
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self._process_name = ""
        self.setUseOpenGL()

    def __getstate__(self):
        return dict(points=self.pointsVector())

    def __setstate__(self, state):
        self.__init__()
        super().append(state["points"])



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtCore
from qtpy.QtCharts import QtCharts

THEMES = {"Light": QtCharts.QChart.ChartThemeLight,
          "Blue Cerulean": QtCharts.QChart.ChartThemeBlueCerulean,
          "Dark": QtCharts.QChart.ChartThemeDark,
          "Brown Sand": QtCharts.QChart.ChartThemeBrownSand,
          "Blue NCS": QtCharts.QChart.ChartThemeBlueNcs,
          "High Contrast": QtCharts.QChart.ChartThemeHighContrast,
          "Blue Icy": QtCharts.QChart.ChartThemeBlueIcy,
          "Qt": QtCharts.QChart.ChartThemeQt}

ALIGNMENTS = dict(left=QtCore.Qt.AlignLeft,
                  right=QtCore.Qt.AlignRight,
                  top=QtCore.Qt.AlignTop,
                  bottom=QtCore.Qt.AlignBottom)

ANIMATION_OPTS = dict(series=QtCharts.QChart.SeriesAnimations)


class Chart(QtCharts.QChart):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_x = 0
        self.max_y = 0
        self.min_x = 0
        self.min_y = 0

    def update_boundaries(self):
        """
        set new min/max values based on axis
        """
        self.max_x = self.axisX().max()
        self.max_y = self.axisY().max()
        self.min_x = self.axisX().min()
        self.min_y = self.axisY().min()

    def hide_legend(self):
        self.legend().hide()

    def show_legend(self):
        self.legend().show()

    def set_legend_alignment(self, alignment: str):
        if alignment not in ALIGNMENTS:
            raise ValueError(f"{alignment!r} not a valid alignment.")
        self.legend().setAlignment(ALIGNMENTS[alignment])

    def set_theme(self, theme_name: str):
        self.setTheme(THEMES[theme_name])

    def set_animation_options(self, option: str):
        self.setAnimationOptions(ANIMATION_OPTS[option])

    def apply_nice_numbers(self):
        """
        adjust both axis to display nice round numbers
        """
        self.axisX().applyNiceNumbers()
        self.axisY().applyNiceNumbers()

    def zoom_by_factor(self, factor: float):
        """
        zoom in/out by factor (1.0 = no change)
        make sure that we dont zoom out too far
        """
        self.zoom(factor)
        if self.axisX().min() < self.min_x:
            self.axisX().setMin(self.min_x)
        if self.axisX().max() > self.max_x:
            self.axisX().setMax(self.max_x)
        if self.axisY().max() > self.max_y:
            self.axisY().setMax(self.max_y)

        # always bottom-align when zooming for now. should perhaps become optional.
        # if self.axisY().min() < self.min_y:
        self.axisY().setMin(max(0, self.min_y))



# -*- coding: utf-8 -*-
"""
"""

import contextlib

from qtpy import QtWidgets

from prettyqt import gui, core, widgets
from prettyqt.utils import colors

QtWidgets.QTextEdit.__bases__ = (widgets.AbstractScrollArea,)


class TextEdit(QtWidgets.QTextEdit):

    value_changed = core.Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.textChanged.connect(self.on_value_change)

    def __getstate__(self):
        return dict(text=self.text(),
                    enabled=self.isEnabled(),
                    font=gui.Font(self.font()))

    def __setstate__(self, state):
        self.__init__()
        self.set_text(state["text"])
        self.setEnabled(state.get("enabled", True))
        self.setFont(state["font"])

    def __add__(self, other):
        if isinstance(other, str):
            self.append_text(other)
            return self

    def on_value_change(self):
        self.value_changed.emit(self.text())

    @contextlib.contextmanager
    def create_cursor(self):
        cursor = gui.TextCursor(self.document())
        yield cursor
        self.setTextCursor(cursor)

    def set_text(self, text: str):
        self.setPlainText(text)

    def append_text(self, text: str):
        self.append(text)

    def text(self) -> str:
        return self.toPlainText()

    def select_text(self, start: int, end: int):
        with self.create_cursor() as c:
            c.select_text(start, end)

    def set_read_only(self, value: bool = True):
        self.setReadOnly(value)

    def set_text_color(self, color: colors.ColorType):
        color = colors.get_color(color)
        self.setTextColor(color)


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    widget = TextEdit("This is a test")
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtWidgets

from prettyqt import widgets


QtWidgets.QInputDialog.__bases__ = (widgets.BaseDialog,)


class InputDialog(QtWidgets.QInputDialog):

    @classmethod
    def get_int(cls, title=None, label=None, icon=None):
        par = widgets.Dialog()
        par.set_icon(icon)
        v = cls.getInt(par, title, label)
        if v[1]:
            return v[0]

    @classmethod
    def get_float(cls, title=None, label=None, icon=None):
        par = widgets.Dialog()
        par.set_icon(icon)
        v = cls.getDouble(par, title, label)
        if v[1]:
            return v[0]

    @classmethod
    def get_text(cls, title=None, label=None, icon=None, default_value=""):
        par = widgets.Dialog()
        par.set_icon(icon)
        v = cls.getText(par, title, label, text=default_value)
        if v[1]:
            return v[0]

    @classmethod
    def get_item(cls, items, title=None, label=None, icon=None):
        par = widgets.Dialog()
        par.set_icon(icon)
        v = cls.getItem(par, title, label, items, editable=False)
        if v[1]:
            return v[0]


if __name__ == "__main__":
    app = widgets.app()
    result = InputDialog.get_text("a", "b", icon="mdi.timer")
    print(result)
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""


from qtpy import QtCore, QtWidgets

from prettyqt import widgets
from prettyqt.utils import bidict

ALLOWED_AREAS = bidict(all=QtCore.Qt.AllDockWidgetAreas,
                       left=QtCore.Qt.LeftDockWidgetArea,
                       right=QtCore.Qt.RightDockWidgetArea,
                       top=QtCore.Qt.TopDockWidgetArea,
                       bottom=QtCore.Qt.BottomDockWidgetArea)


QtWidgets.QDockWidget.__bases__ = (widgets.Widget,)


class DockWidget(QtWidgets.QDockWidget):
    """
    Customized DockWidget class
    contains a custom TitleBar with maximise button
    """

    def __init__(self, *args, **kwargs):
        name = kwargs.pop("name", None)
        title = kwargs.pop("title", None)
        super().__init__(*args, **kwargs)
        if name:
            self.id = name
        if title:
            self.title = title
        self.set_allowed_areas("all")

    def set_widget(self, widget):
        self.setWidget(widget)

    def set_allowed_areas(self, area):
        self.setAllowedAreas(ALLOWED_AREAS[area])

    def setup_title_bar(self):
        title_bar = widgets.Widget()
        layout = widgets.BoxLayout("horizontal")
        layout.set_margin(0)
        layout.set_alignment("right")
        title_bar.set_layout(layout)
        maximise_button = widgets.PushButton()
        layout += maximise_button
        maximise_button.set_style_icon("maximise")
        maximise_button.clicked.connect(self.maximise)
        close_button = widgets.PushButton()
        close_button.set_style_icon("close")
        layout += close_button
        close_button.clicked.connect(self.close)
        self.setTitleBarWidget(title_bar)

    def maximise(self):
        if not self.isFloating():
            self.setFloating(True)
        if not self.isMaximized():
            self.showMaximized()
        else:
            self.showMinimized()


if __name__ == "__main__":
    app = widgets.app()
    win = widgets.MainWindow()
    dock_widget = DockWidget(name="aa", title="Test")
    dock_widget.setup_title_bar()
    win.add_dockwidget(dock_widget, "left")
    win.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""

import contextlib

from qtpy import QtCore

from prettyqt import core

OPEN_MODES = core.iodevice.OPEN_MODES

QtCore.QFile.__bases__ = (core.FileDevice,)


class File(QtCore.QFile):

    @contextlib.contextmanager
    def open_file(self, flag):
        if flag in OPEN_MODES:
            flag = OPEN_MODES[flag]
        self.open(flag)
        yield None
        self.close()



# -*- coding: utf-8 -*-

"""gui module

contains QWebEngineView-based classes
"""

import logging

try:
    from .webengineview import WebEngineView
    from .webenginepage import WebEnginePage
    __all__ = ["WebEngineView", "WebEnginePage"]
except ModuleNotFoundError:
    logging.warning("Tried to import WebEngineView but not installed.")



# -*- coding: utf-8 -*-
"""
"""

import functools
from typing import Iterable, Optional

from qtpy import QtCore, QtWidgets

from prettyqt import core, widgets
from prettyqt.utils import bidict

QtWidgets.QHeaderView.__bases__ = (widgets.AbstractItemView,)


class HeaderView(QtWidgets.QHeaderView):
    MODES = bidict(interactive=QtWidgets.QHeaderView.Interactive,
                   fixed=QtWidgets.QHeaderView.Fixed,
                   stretch=QtWidgets.QHeaderView.Stretch,
                   resize_to_contents=QtWidgets.QHeaderView.ResizeToContents)

    section_vis_changed = QtCore.Signal(int, bool)

    def __init__(self, orientation=None, parent=None):
        o = QtCore.Qt.Vertical if orientation == "vertical" else QtCore.Qt.Horizontal
        super().__init__(o, parent=parent)
        self.setSectionsMovable(True)
        self.setSectionsClickable(True)
        self.widget_name = parent.id

    def save_state(self):
        settings = core.Settings()
        settings.setValue(f"{self.widget_name}.state", self.saveState())

    def load_state(self):
        settings = core.Settings()
        state = settings.value(f"{self.widget_name}.state", None)
        if state is not None:
            self.restoreState(state)

    def resize_sections(self, mode: str):
        self.resizeSections(self.MODES[mode])

    def resize_mode(self, mode: str, col: Optional[int] = None):
        if mode not in self.MODES:
            raise ValueError("mode not existing")
        if col is None:
            self.setSectionResizeMode(self.MODES[mode])
        else:
            self.setSectionResizeMode(col, self.MODES[mode])

    def section_labels(self) -> list:
        model = self.parent().model()
        return [model.headerData(i, QtCore.Qt.Horizontal, QtCore.Qt.DisplayRole)
                for i in range(self.count())]

    def contextMenuEvent(self, event):
        """
        context menu for our files tree
        """
        menu = widgets.Menu(parent=self)
        for i, header_label in enumerate(self.section_labels()[1:], start=1):
            act = menu.addAction(header_label)
            act.setCheckable(True)
            val = not self.isSectionHidden(i)
            act.setChecked(val)
            fn = functools.partial(self.set_section_hidden, i=i, hide=val)
            act.triggered.connect(fn)
        menu.exec_(self.mapToGlobal(event.pos()))

    def set_section_hidden(self, i: int, hide: bool):
        self.section_vis_changed.emit(i, hide)
        self.setSectionHidden(i, hide)

    def set_sizes(self, sizes: Iterable):
        for i, size in enumerate(sizes):
            if size is not None:
                self.resizeSection(i, size)

    def set_default_section_size(self, size):
        if size is None:
            self.resetDefaultSectionSize()
        else:
            self.setDefaultSectionSize(size)

    def stretch_last_section(self, stretch: bool = True):
        self.setStretchLastSection(stretch)


if __name__ == "__main__":
    app = widgets.app()
    header = HeaderView(parent=None)
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtWidgets

from prettyqt import widgets, core


QtWidgets.QCommandLinkButton.__bases__ = (widgets.PushButton,)


class CommandLinkButton(QtWidgets.QCommandLinkButton):

    value_changed = core.Signal(bool)


if __name__ == "__main__":
    app = widgets.app()
    widget = CommandLinkButton("This is a test")
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""

import pathlib

import qtawesome as qta
from qtpy import QtCore, QtGui

from prettyqt import gui


class Icon(QtGui.QIcon):

    def __init__(self, icon=None):
        if isinstance(icon, pathlib.Path):
            icon = str(icon)
        super().__init__(icon)

    # def __reduce__(self):
    #     return type(self), (), self.__getstate__()

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __bool__(self):
        return not bool(self.isNull())

    def __getstate__(self):
        ba = QtCore.QByteArray()
        stream = QtCore.QDataStream(ba, QtCore.QIODevice.WriteOnly)
        pixmap = self.pixmap(QtCore.QSize(256, 256))
        stream << pixmap
        return ba

    def __setstate__(self, ba):
        stream = QtCore.QDataStream(ba, QtCore.QIODevice.ReadOnly)
        px = QtGui.QPixmap()
        stream >> px
        super().__init__(px)

    @staticmethod
    def by_name(name: str):
        return qta.icon(name)

    @classmethod
    def for_color(cls, color_str: str):
        color = gui.Color.from_text(color_str)
        if color.isValid():
            bitmap = gui.Pixmap(16, 16)
            bitmap.fill(color)
            icon = gui.Icon(bitmap)
        else:
            icon = qta.icon("mdi.card-outline")
        return icon


if __name__ == "__main__":
    from prettyqt import widgets
    app = widgets.app()
    icon = Icon.for_color("green")



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtWidgets


class ToolTip(QtWidgets.QToolTip):

    def set_formatted_text(self, text, linebreak_px=400):
        self.setText(f'<div style="max-width: {linebreak_px}px">{text}</div>')



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtCore, QtWidgets

from prettyqt import widgets
from prettyqt.utils import bidict

area = QtWidgets.QAbstractScrollArea

SIZE_POLICIES = bidict(content=area.AdjustToContents,
                       first_show=area.AdjustToContentsOnFirstShow,
                       ignored=area.AdjustIgnored)


SCROLLBAR_POLICY = bidict(always_on=QtCore.Qt.ScrollBarAlwaysOn,
                          always_off=QtCore.Qt.ScrollBarAlwaysOff,
                          as_needed=QtCore.Qt.ScrollBarAsNeeded)


QtWidgets.QAbstractScrollArea.__bases__ = (widgets.Frame,)


class AbstractScrollArea(QtWidgets.QAbstractScrollArea):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.h_scrollbar = widgets.ScrollBar()
        self.v_scrollbar = widgets.ScrollBar()
        self.setHorizontalScrollBar(self.h_scrollbar)
        self.setVerticalScrollBar(self.v_scrollbar)

    # @property
    # def h_scrollbar(self):
    #     return self.horizontalScrollbar()

    # @h_scrollbar.setter
    # def h_scrollbar(self, scrollbar):
    #     self.setHorizontalScrollBar(scrollbar)

    # @property
    # def v_scrollbar(self):
    #     return self.verticalScrollBar()

    # @v_scrollbar.setter
    # def v_scrollbar(self, scrollbar):
    #     self.setVerticalScrollBar(scrollbar)

    def set_size_adjust_policy(self, policy: str):
        """set size adjust policy

        Valid values are "content", "first_show", "ignored"

        Args:
            policy: size adjust policy to use

        Raises:
            ValueError: invalid size adjust policy
        """
        if policy not in SIZE_POLICIES:
            raise ValueError("Policy not available")
        policy = SIZE_POLICIES.get(policy)
        self.setSizeAdjustPolicy(policy)

    def get_size_adjust_policy(self) -> str:
        """returns size adjust policy

        possible values are "content", "first_show", "ignored"

        Returns:
            size adjust policy
        """
        return SIZE_POLICIES.inv[self.sizeAdjustPolicy()]

    def set_scrollbar_policy(self, mode: str):
        """sets the policy for both scrollbars

        possible values are "always_on", "always_off", "as_needed"

        Args:
            mode: visibilty to set

        Raises:
            ValueError: invalid scrollbar policy
        """
        if mode not in SCROLLBAR_POLICY:
            raise ValueError("Invalid scrollbar policy")
        self.setHorizontalScrollBarPolicy(SCROLLBAR_POLICY[mode])
        self.setVerticalScrollBarPolicy(SCROLLBAR_POLICY[mode])

    def set_horizontal_scrollbar_policy(self, mode: str):
        """sets the horizontal scrollbar visibility

        possible values are "always_on", "always_off", "as_needed"

        Args:
            mode: visibilty to set

        Raises:
            ValueError: invalid scrollbar policy
        """
        if mode not in SCROLLBAR_POLICY:
            raise ValueError("Invalid scrollbar policy")
        self.setHorizontalScrollBarPolicy(SCROLLBAR_POLICY[mode])

    def set_vertical_scrollbar_policy(self, mode: str):
        """sets the vertical scrollbar visibility

        possible values are "always_on", "always_off", "as_needed"

        Args:
            mode: visibilty to set

        Raises:
            ValueError: invalid scrollbar policy
        """
        if mode not in SCROLLBAR_POLICY:
            raise ValueError("Invalid scrollbar policy")
        self.setVerticalScrollBarPolicy(SCROLLBAR_POLICY[mode])

    def set_scrollbar_width(self, width: int):
        """sets the width for both scrollbars

        Args:
            width: width in pixels
        """
        self.set_horizontal_scrollbar_width(width)
        self.set_vertical_scrollbar_width(width)

    def set_horizontal_scrollbar_width(self, width: int):
        """sets the horizontal scrollbar width

        Args:
            width: width in pixels
        """
        stylesheet = f"QScrollBar:horizontal {{height: {width}px;}}"
        self.horizontalScrollBar().setStyleSheet(stylesheet)

    def set_vertical_scrollbar_width(self, width: int):
        """sets the vertical scrollbar width

        Args:
            width: width in pixels
        """
        stylesheet = f"QScrollBar:vertical {{width: {width}px;}}"
        self.verticalScrollBar().setStyleSheet(stylesheet)

    def scroll_to_top(self):
        """scroll to the top of the scroll area
        """
        self.verticalScrollBar().scroll_to_min()

    def scroll_to_bottom(self):
        """scroll to the bottom of the scroll area
        """
        self.verticalScrollBar().scroll_to_max()


if __name__ == "__main__":
    app = widgets.app()
    widget = AbstractScrollArea()
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtWidgets

from prettyqt import core, widgets

QtWidgets.QSpinBox.__bases__ = (widgets.AbstractSpinBox,)


class SpinBox(QtWidgets.QSpinBox):

    value_changed = core.Signal(int)

    def __init__(self, parent=None, min_value=None, max_value=None, default_value=None):
        super().__init__(parent)
        self.valueChanged.connect(self.value_changed)
        self.set_range(min_value, max_value)
        if default_value is not None:
            self.set_value(default_value)

    def __getstate__(self):
        return dict(range=(self.minimum(), self.maximum()),
                    value=self.value(),
                    enabled=self.isEnabled(),
                    tooltip=self.toolTip(),
                    statustip=self.statusTip(),
                    prefix=self.prefix(),
                    suffix=self.suffix(),
                    int_base=self.displayIntegerBase(),
                    step_type=self.get_step_type(),
                    button_symbols=self.get_button_symbols(),
                    correction_mode=self.get_correction_mode(),
                    single_step=self.singleStep())

    def __setstate__(self, state):
        self.__init__()
        self.set_range(*state["range"])
        self.setValue(state["value"])
        self.setEnabled(state.get("enabled", True))
        self.setToolTip(state.get("tooltip", ""))
        self.setStatusTip(state.get("statustip", ""))
        self.setSingleStep(state["single_step"])
        self.setPrefix(state["prefix"])
        self.setSuffix(state["suffix"])
        self.set_button_symbols(state["button_symbols"])
        self.set_correction_mode(state["correction_mode"])
        self.setDisplayIntegerBase(state["int_base"])
        self.set_step_type(state["step_type"])

    def set_range(self, start, end):
        if start is None:
            start = -2147483647
        if end is None:
            end = 2147483647
        self.setRange(start, end)

    def set_step_size(self, step_size):
        self.setSingleStep(step_size)


if __name__ == "__main__":
    app = widgets.app()
    widget = SpinBox()
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""

for full list, see:
- https://cdn.materialdesignicons.com/3.0.39/
"""

from typing import Union
import pathlib

import qtawesome as qta
from qtpy import QtGui
from prettyqt import gui


IconType = Union[QtGui.QIcon, str, pathlib.Path, None]


def get_icon(icon: IconType, color: str = "black"):
    if icon is None:
        icon = gui.Icon()
    elif isinstance(icon, str):
        if icon.startswith("mdi."):
            icon = qta.icon(icon, color=color)
        else:
            icon = gui.Icon(icon)
    elif isinstance(icon, pathlib.Path):
        icon = gui.Icon(icon)
    return icon


def set_defaults(*args, **kwargs):
    qta.set_defaults(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
"""

import pathlib

from qtpy import QtCore


class Url(QtCore.QUrl):

    # def __str__(self):
    #     return self.absolutePath()

    def to_path(self) -> pathlib.Path:
        return pathlib.Path(self.path())

    def is_local_file(self) -> bool:
        return self.isLocalFile()



# -*- coding: utf-8 -*-
"""
"""
# see https://github.com/ITVRoC/SeekurJr/blob/master/seekur_12.04/packages/
# multimaster_fkie/node_manager_fkie/src/node_manager_fkie/yaml_highlighter.py

from dataclasses import dataclass
import regex as re
from prettyqt import gui


@dataclass
class HighlightRule(object):
    regex = ""
    color = "black"
    italic = False
    bold = False
    minimal = False
    font_size = None
    nth = 0

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if isinstance(cls.regex, str):
            cls.compiled = re.compile(cls.regex)
            # cls.compiled.setMinimal(True)
        else:
            cls.compiled = [re.compile(r) for r in cls.regex]
        cls.format = cls.get_format()

    @classmethod
    def get_format(cls):
        fmt = gui.TextCharFormat()
        fmt.setFontItalic(cls.italic)
        fmt.set_foreground_color(cls.color)
        if cls.font_size:
            fmt.setFontPointSize(cls.font_size)
        if cls.bold:
            fmt.set_font_weight("bold")
        return fmt



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtCore

from prettyqt import gui, widgets


class IconLabel(widgets.Widget):

    def __init__(self, text=None,
                 tooltip="",
                 icon="mdi.help-circle-outline",
                 parent=None):
        super().__init__(parent=parent)
        self.set_layout("horizontal")
        self.label = widgets.Label(text)
        self.label.setMargin(10)
        self.label.set_size_policy(horizontal="minimum")
        self.tooltip = tooltip
        icon = gui.Icon.by_name(icon)
        self.icon = widgets.Label()
        self.icon.setToolTip(tooltip)
        self.icon.set_size_policy(horizontal="minimum")
        pixmap = icon.pixmap(QtCore.QSize(20, 20))
        self.icon.setPixmap(pixmap)
        self.box += self.label
        self.box += self.icon
        self.box.setSpacing(0)
        self.box.addStretch()

    def __getattr__(self, value):
        return self.label.__getattribute__(value)

    def __repr__(self):
        return f"IconLabel({self.text()!r})"


if __name__ == "__main__":
    app = widgets.app()
    widget = IconLabel("test", tooltip="testus")
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtWidgets, QtCore

from prettyqt import widgets, core
from prettyqt.utils import bidict

TICK_POSITIONS = bidict(none=QtWidgets.QSlider.NoTicks,
                        both_sides=QtWidgets.QSlider.TicksBothSides,
                        above=QtWidgets.QSlider.TicksAbove,
                        below=QtWidgets.QSlider.TicksBelow)

ORIENTATIONS = bidict(horizontal=QtCore.Qt.Horizontal,
                      vertical=QtCore.Qt.Vertical)


QtWidgets.QSlider.__bases__ = (widgets.AbstractSlider,)


class Slider(QtWidgets.QSlider):

    value_changed = core.Signal(int)

    def __init__(self, orientation="horizontal", parent=None):
        if orientation in ORIENTATIONS:
            orientation = ORIENTATIONS[orientation]
        super().__init__(orientation, parent)
        self.valueChanged.connect(self.on_value_change)

    def __getstate__(self):
        return dict(range=(self.minimum(), self.maximum()),
                    value=self.value(),
                    tooltip=self.toolTip(),
                    statustip=self.statusTip(),
                    enabled=self.isEnabled(),
                    has_tracking=self.hasTracking(),
                    tick_position=self.get_tick_position(),
                    tick_interval=self.tickInterval(),
                    inverted_controls=self.invertedControls(),
                    inverted_appearance=self.invertedAppearance(),
                    single_step=self.singleStep(),
                    page_step=self.pageStep())

    def __setstate__(self, state):
        self.__init__()
        self.set_range(*state["range"])
        self.set_value(state["value"])
        self.setToolTip(state.get("tooltip", ""))
        self.setStatusTip(state.get("statustip", ""))
        self.setEnabled(state.get("enabled", True))
        self.setSingleStep(state["single_step"])
        self.setPageStep(state["page_step"])
        self.setTracking(state["has_tracking"])
        self.setInvertedControls(state["inverted_controls"])
        self.setInvertedAppearance(state["inverted_appearance"])
        self.set_tick_position(state["tick_position"])
        self.setTickInterval(state["tick_interval"])

    def set_tick_position(self, position: str):
        """sets the tick position for the slider

        allowed values are "none", "both_sides", "above", "below", "left", "right"
        for vertical orientation of the slider,
        "above" equals to "left" and "below" to "right"

        Args:
            position: position for the ticks
        """
        if position == "left":
            position = "above"
        elif position == "right":
            position = "below"
        elif position not in TICK_POSITIONS:
            raise ValueError(f"{position} not a valid tick position.")
        self.setTickPosition(TICK_POSITIONS[position])

    def get_tick_position(self) -> str:
        """returns tick position

        possible values are "none", "both_sides", "above", "below"

        Returns:
            tick position
        """
        val = TICK_POSITIONS.inv[self.tickPosition()]
        # if self.is_vertical():
        #     if val == "above":
        #         return "left"
        #     elif val == "below":
        #         return "right"
        return val


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    slider = Slider()
    slider.setRange(0, 100)
    slider.value_changed.connect(print)
    slider.show()
    app.exec_()



# -*- coding: utf-8 -*-

"""widgets module

contains QtWidgets-based classes
"""

from .application import Application
from .widget import Widget
from .frame import Frame
from .abstractslider import AbstractSlider
from .abstractscrollarea import AbstractScrollArea
from .abstractbutton import AbstractButton
from .abstractspinbox import AbstractSpinBox
from .abstractitemview import AbstractItemView
from .mdisubwindow import MdiSubWindow
from .mdiarea import MdiArea
from .scrollbar import ScrollBar
from .scrollarea import ScrollArea
from .desktopwidget import DesktopWidget
from .graphicsitem import GraphicsItem
from .style import Style
from .styleoption import StyleOption
from .sizepolicy import SizePolicy
from .stylepainter import StylePainter
from .dialog import BaseDialog, Dialog
from .messagebox import MessageBox
from .errormessage import ErrorMessage

from .filesystemmodel import FileSystemModel

from .slider import Slider
from .styleoptionslider import StyleOptionSlider
from .styleoptionframe import StyleOptionFrame

from .dockwidget import DockWidget

from .mainwindow import MainWindow

from .listwidgetitem import ListWidgetItem
from .treewidgetitem import TreeWidgetItem
from .action import Action
from .widgetaction import WidgetAction
from .toolbutton import ToolButton
from .tooltip import ToolTip
from .menu import Menu
from .menubar import MenuBar
from .statusbar import StatusBar
from .tabbar import TabBar
from .tabwidget import TabWidget
from .toolbar import ToolBar
from .headerview import HeaderView
from .label import Label
from .pushbutton import PushButton
from .commandlinkbutton import CommandLinkButton
from .radiobutton import RadioButton
from .combobox import ComboBox
from .spinbox import SpinBox
from .doublespinbox import DoubleSpinBox
from .checkbox import CheckBox
from .lineedit import LineEdit
from .keysequenceedit import KeySequenceEdit
from .textedit import TextEdit
from .datetimeedit import DateTimeEdit
from .dateedit import DateEdit
from .timeedit import TimeEdit
from .calendarwidget import CalendarWidget
from .plaintextedit import PlainTextEdit
from .textbrowser import TextBrowser
from .progressbar import ProgressBar
from .columnview import ColumnView
from .listview import ListView
from .listwidget import ListWidget
from .treeview import TreeView
from .treewidget import TreeWidget
from .tableview import TableView
from .tablewidgetitem import TableWidgetItem
from .tablewidget import TableWidget
from .abstractitemdelegate import AbstractItemDelegate
from .itemdelegate import ItemDelegate
from .styleditemdelegate import StyledItemDelegate

from .layoutitem import LayoutItem
from .layout import Layout
from .spaceritem import SpacerItem
from .formlayout import FormLayout
from .boxlayout import BoxLayout
from .stackedlayout import StackedLayout
from .gridlayout import GridLayout
from .toolbox import ToolBox

from .splashscreen import SplashScreen
from .progressdialog import ProgressDialog
from .fontdialog import FontDialog
from .filedialog import FileDialog
from .colordialog import ColorDialog
from .inputdialog import InputDialog
from .dialogbuttonbox import DialogButtonBox
from .buttongroup import ButtonGroup
from .groupbox import GroupBox
from .splitter import Splitter
from .wizard import Wizard
from .wizardpage import WizardPage


def app():
    if Application.instance() is not None:
        return Application.instance()
    return Application([])


__all__ = ["app",
           "Application",
           "AbstractSlider",
           "AbstractButton",
           "AbstractSpinBox",
           "AbstractScrollArea",
           "AbstractItemView",
           "MdiSubWindow",
           "MdiArea",
           "ScrollBar",
           "ScrollArea",
           "Widget",
           "DesktopWidget",
           "GraphicsItem",
           "Style",
           "StyleOption",
           "SpacerItem",
           "SizePolicy",
           "StylePainter",
           "BaseDialog",
           "Dialog",
           "MessageBox",
           "ErrorMessage",
           "FileSystemModel",
           "LayoutItem",
           "Layout",
           "FormLayout",
           "BoxLayout",
           "StackedLayout",
           "GridLayout",
           "ToolBox",
           "Slider",
           "StyleOptionSlider",
           "StyleOptionFrame",
           "Frame",
           "ListWidgetItem",
           "TreeWidgetItem",
           "Action",
           "WidgetAction",
           "ToolButton",
           "ToolTip",
           "Menu",
           "MenuBar",
           "StatusBar",
           "TabWidget",
           "TabBar",
           "ToolBar",
           "HeaderView",
           "DockWidget",
           "Label",
           "PushButton",
           "CommandLinkButton",
           "RadioButton",
           "ComboBox",
           "SpinBox",
           "DoubleSpinBox",
           "CheckBox",
           "LineEdit",
           "KeySequenceEdit",
           "TextEdit",
           "DateEdit",
           "TimeEdit",
           "DateTimeEdit",
           "CalendarWidget",
           "PlainTextEdit",
           "TextBrowser",
           "ProgressBar",
           "ColumnView",
           "ListView",
           "ListWidget",
           "TreeView",
           "TreeWidget",
           "TableView",
           "TableWidgetItem",
           "TableWidget",
           "SplashScreen",
           "ProgressDialog",
           "FontDialog",
           "FileDialog",
           "ColorDialog",
           "InputDialog",
           "DialogButtonBox",
           "ButtonGroup",
           "GroupBox",
           "Splitter",
           "Wizard",
           "WizardPage",
           "MainWindow",
           "AbstractItemDelegate",
           "ItemDelegate",
           "StyledItemDelegate"]



# -*- coding: utf-8 -*-
"""
"""

# based on https://stackoverflow.com/questions/58891116/delegate-with-radio-buttons
# credits to musicamante


from qtpy import QtCore, QtWidgets

from prettyqt import core, gui, widgets


class RadioDelegate(widgets.StyledItemDelegate):
    def __init__(self, owner, chs):
        super().__init__(owner)
        self.items = chs
        self.choices = [None for i in self.items]

    def createEditor(self, parent, option, index):
        editor = widgets.Widget(parent)
        editor.setContentsMargins(0, 0, 0, 0)
        editor.setAutoFillBackground(True)
        # create a button group to keep track of the checked radio
        editor.buttonGroup = widgets.ButtonGroup()
        # adding the widget as an argument to the layout constructor automatically
        # applies it to the widget
        layout = widgets.BoxLayout("horizontal", parent=editor)
        layout.setContentsMargins(0, 0, 0, 0)
        for i, k in enumerate(self.items):
            rb = widgets.RadioButton(k)
            layout.addWidget(rb)
            # prevent the radio to get focus from keyboard or mouse
            rb.setFocusPolicy(QtCore.Qt.NoFocus)
            rb.installEventFilter(self)
            editor.buttonGroup.addButton(rb, i)
        # add a stretch to always align contents to the left
        layout.addStretch(1)

        # set a property that will be used for the mask
        editor.setProperty('offMask', gui.Region(editor.rect()))
        editor.installEventFilter(self)
        return editor

    def eventFilter(self, source, event):
        if event.type() == core.Event.MouseButtonPress:
            if isinstance(source, QtWidgets.QRadioButton):
                if not source.parent().hasFocus():
                    # the parent has no focus, set it and ignore the click
                    source.parent().setFocus()
                    return True
            elif not source.hasFocus():
                # the container has been clicked, check
                source.setFocus()
        elif event.type() == core.Event.FocusIn:
            # event received as a consequence of setFocus
            # clear the mask to show it completely
            source.clearMask()
        elif event.type() == core.Event.FocusOut:
            # another widget has requested focus, set the mask
            source.setMask(source.property('offMask'))
            # update the table viewport to get rid of possible
            # grid lines left after masking
            source.parent().update()
        return super().eventFilter(source, event)

    def updateEditorGeometry(self, editor, option, index):
        rect = core.Rect(option.rect)
        minWidth = editor.minimumSizeHint().width()
        if rect.width() < minWidth:
            rect.setWidth(minWidth)
        editor.setGeometry(rect)
        # create a new mask based on the option rectangle, then apply it
        mask = gui.Region(0, 0, option.rect.width(), option.rect.height())
        editor.setProperty('offMask', mask)
        editor.setMask(mask)

    def setEditorData(self, editor, index):
        value = index.data(QtCore.Qt.DisplayRole)
        if value in self.items:
            editor.buttonGroup.button(self.items.index(value)).setChecked(True)

    def setModelData(self, editor, model, index):
        button = editor.buttonGroup.checkedId()
        if button >= 0:
            model.setData(index, self.items[button], QtCore.Qt.DisplayRole)
            self.choices[button] = index.row()


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    widget = widgets.TableWidget()
    widget.setColumnCount(3)
    widget.insertRow(0)
    widget.insertRow(0)
    widget.setHorizontalHeaderLabels(['LIB', 'CELL', 'area'])
    item = widgets.TableWidgetItem("test")
    widget.setItem(0, 0, item)
    widget.setItem(1, 1, widgets.TableWidgetItem("test"))
    widget.setItem(1, 2, widgets.TableWidgetItem("test"))
    widget.setItem(2, 1, widgets.TableWidgetItem("test"))
    delegate = RadioDelegate(widget, ["a", "b"])
    widget.setItemDelegateForColumn(0, delegate)
    widget.openPersistentEditor(item)
    widget.show()
    app.exec_()
    print(delegate.choices)



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtWidgets

from prettyqt import widgets


QtWidgets.QMdiSubWindow.__bases__ = (widgets.Widget,)


class MdiSubWindow(QtWidgets.QMdiSubWindow):
    pass


if __name__ == "__main__":
    app = widgets.app()
    widget = MdiSubWindow()
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtCore, QtWidgets
from qtpy.QtCharts import QtCharts

from prettyqt import charts, core, gui, widgets

ALIGNMENTS = dict(left=QtCore.Qt.AlignLeft,
                  right=QtCore.Qt.AlignRight,
                  top=QtCore.Qt.AlignTop,
                  bottom=QtCore.Qt.AlignBottom)

ZOOM_IN_FACTOR = 1.1
ZOOM_OUT_FACTOR = 1.0 / ZOOM_IN_FACTOR
SCROLL_STEP_SIZE = 10


class ChartView(QtCharts.QChartView):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        chart = charts.Chart()
        self.setChart(chart)
        self.setRenderHint(gui.Painter.Antialiasing)
        self.setRubberBand(self.RectangleRubberBand)
        # self.setDragMode(self.RubberBandDrag)

    def keyPressEvent(self, event):
        """
        handle keypress events to allow navigation via keyboard
        """
        key = event.key()
        if key == QtCore.Qt.Key_Escape:
            self.chart().zoomReset()
        elif key == QtCore.Qt.Key_Plus:
            self.chart().zoom_by_factor(ZOOM_IN_FACTOR)
        elif key == QtCore.Qt.Key_Minus:
            self.chart().zoom_by_factor(ZOOM_OUT_FACTOR)
        elif key == QtCore.Qt.Key_Left:
            self.chart().scroll(-SCROLL_STEP_SIZE, 0)
        elif key == QtCore.Qt.Key_Right:
            self.chart().scroll(SCROLL_STEP_SIZE, 0)
        elif key == QtCore.Qt.Key_Up:
            self.chart().scroll(0, SCROLL_STEP_SIZE)
        elif key == QtCore.Qt.Key_Down:
            self.chart().scroll(0, -SCROLL_STEP_SIZE)
        elif key == QtCore.Qt.Key_0:
            self.chart().apply_nice_numbers()
        else:
            return super().keyPressEvent(event)
        event.accept()

    def wheelEvent(self, event):
        """
        handle wheel event for zooming
        """
        fct = ZOOM_IN_FACTOR if event.angleDelta().y() > 0 else ZOOM_OUT_FACTOR
        self.chart().zoom_by_factor(fct)

    def mouseReleaseEvent(self, event):
        """
        override to allow dragging the chart
        """
        if event.button() == QtCore.Qt.RightButton:
            QtWidgets.QApplication.restoreOverrideCursor()
            event.accept()
            return None
        super().mouseReleaseEvent(event)

    def mousePressEvent(self, event):
        """
        override to allow dragging the chart
        """
        if event.button() == QtCore.Qt.RightButton:
            cursor = gui.Cursor(QtCore.Qt.SizeAllCursor)
            QtWidgets.QApplication.setOverrideCursor(cursor)
            self.last_mouse_pos = event.pos()
            event.accept()

        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        """
        override to allow dragging the chart
        """
        # pan the chart with a middle mouse drag
        if event.buttons() & QtCore.Qt.RightButton:
            if not self.last_mouse_pos:
                return None
            pos_diff = event.pos() - self.last_mouse_pos
            self.chart().scroll(-pos_diff.x(), pos_diff.y())

            self.last_mouse_pos = event.pos()
            event.accept()

            QtWidgets.QApplication.restoreOverrideCursor()

        super().mouseMoveEvent(event)

    @core.Slot()
    def save(self):
        """
        let user choose folder and save chart as an image file
        """
        dlg = widgets.FileDialog(mode="save", caption="Save image")
        filters = {"Bmp files": [".bmp"],
                   "Jpeg files": [".jpg"],
                   "Png files": [".png"]}
        dlg.set_filter(filters)
        filename = dlg.open_file()
        if not filename:
            return None
        self.chart().show_legend()
        image = self.get_image()
        image.save(filename[0])
        self.chart().hide_legend()

    def get_image(self):
        image = self.grab()
        gl_widget = self.findChild(QtWidgets.QOpenGLWidget)
        if gl_widget:
            d = gl_widget.mapToGlobal(core.Point()) - self.mapToGlobal(core.Point())
            with gui.Painter(image) as painter:
                painter.set_composition_mode("source_atop")
                painter.drawImage(d, gl_widget.grabFramebuffer())
        return image

    # def hide_legend(self):
    #     self.chart().hide_legend()

    # def show_legend(self):
    #     self.chart().show_legend()

    # def set_legend_alignment(self, alignment: str):
    #     if alignment not in ALIGNMENTS:
    #         raise ValueError(f"{alignment!r} not a valid alignment.")
    #     self.chart().legend().setAlignment(ALIGNMENTS[alignment])



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtGui

from prettyqt import gui
from prettyqt.utils import bidict


STYLE_HINTS = bidict(any=QtGui.QFont.AnyStyle,
                     sans_serif=QtGui.QFont.SansSerif,
                     serif=QtGui.QFont.Serif,
                     typewriter=QtGui.QFont.TypeWriter,
                     decorative=QtGui.QFont.Decorative,
                     monospace=QtGui.QFont.Monospace,
                     fantasy=QtGui.QFont.Fantasy,
                     cursive=QtGui.QFont.Cursive,
                     system=QtGui.QFont.System)


class Font(QtGui.QFont):

    def __repr__(self):
        return (f"Font({self.family()}, {self.pointSize()}, "
                f"{self.weight()}, {self.italic()})")

    def __getstate__(self):
        return dict(family=self.family(),
                    pointsize=self.pointSize(),
                    weight=self.weight(),
                    italic=self.italic())

    def __setstate__(self, state):
        self.__init__()
        self.setFamily(state["family"])
        if state["pointsize"] > -1:
            self.setPointSize(state["pointsize"])
        self.setWeight(state["weight"])
        self.setItalic(state["italic"])

    @property
    def metrics(self):
        return gui.FontMetrics(self)

    def set_size(self, size: int):
        pass

    @classmethod
    def mono(cls, size=8):
        font = cls("Consolas", size)
        # font.setStyleHint()
        return font

    def set_style_hint(self, hint: str):
        """sets the style hint

        Valid values are "any", "sans_serif", "serif", "typewriter", "decorative",
        "monospace", "fantasy", "cursive", "system"

        Args:
            hint: style hint

        Raises:
            ValueError: invalid style hint
        """
        if hint not in STYLE_HINTS:
            raise ValueError("Invalid style hint")
        self.setStyleHint(STYLE_HINTS[hint])


if __name__ == "__main__":
    font = Font("Consolas")



# -*- coding: utf-8 -*-
"""custom_widgets module
"""

from .image import Image
from .listinput import ListInput
from .booldicttoolbutton import BoolDictToolButton
from .optionalwidget import OptionalWidget
from .singlelinetextedit import SingleLineTextEdit
from .regexinput import RegexInput
from .mappedcheckbox import MappedCheckBox
from .promptlineedit import PromptLineEdit
from .logtextedit import LogTextEdit
from .flagselectionwidget import FlagSelectionWidget
from .stringornumberwidget import StringOrNumberWidget
from .iconlabel import IconLabel
from .flowlayout import FlowLayout
from .sidebarwidget import SidebarWidget
from .colorchooserbutton import ColorChooserButton
from .filechooserbutton import FileChooserButton
from .fontchooserbutton import FontChooserButton
from .inputandslider import InputAndSlider
from .spanslider import SpanSlider
from .labeledslider import LabeledSlider
from .waitingspinner import WaitingSpinner
from .markdownwidget import MarkdownWindow
from .imageviewer import ImageViewer
from .popupinfo import PopupInfo
from .buttondelegate import ButtonDelegate
from .radiodelegate import RadioDelegate
from .selectionwidget import SelectionWidget
from .codeeditor import CodeEditor
from .regexeditor.editor import RegexEditorWidget

__all__ = ["Image",
           "ListInput",
           "BoolDictToolButton",
           "OptionalWidget",
           "SingleLineTextEdit",
           "RegexInput",
           "MappedCheckBox",
           "PromptLineEdit",
           "LogTextEdit",
           "FlagSelectionWidget",
           "StringOrNumberWidget",
           "IconLabel",
           "FlowLayout",
           "SidebarWidget",
           "ColorChooserButton",
           "FileChooserButton",
           "FontChooserButton",
           "InputAndSlider",
           "SpanSlider",
           "LabeledSlider",
           "WaitingSpinner",
           "PopupInfo",
           "ButtonDelegate",
           "RadioDelegate",
           "SelectionWidget",
           "ImageViewer",
           "MarkdownWindow",
           "CodeEditor",
           "RegexEditorWidget"]



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtWidgets

from prettyqt import widgets


QtWidgets.QSpacerItem.__bases__ = (widgets.LayoutItem,)


class SpacerItem(QtWidgets.QSpacerItem):
    pass



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtGui, QtCore
from prettyqt.utils import bidict

ELIDE_MODES = bidict(left=QtCore.Qt.ElideLeft,
                     right=QtCore.Qt.ElideRight,
                     middle=QtCore.Qt.ElideMiddle,
                     none=QtCore.Qt.ElideNone)


class FontMetrics(QtGui.QFontMetrics):

    def elided_text(self, text: str, mode: str, width: int, flags=0) -> str:
        if mode not in ELIDE_MODES:
            raise ValueError("Mode not available")
        return self.elidedText(text, ELIDE_MODES[mode], width, flags)



# -*- coding: utf-8 -*-
"""
"""

from prettyqt import core, gui, widgets


class FontChooserButton(widgets.Widget):

    value_changed = core.Signal(gui.Font)

    def __init__(self, font=None, parent=None):
        super().__init__(parent)
        self.current_font = font
        layout = widgets.BoxLayout("horizontal", self)
        layout.set_margin(0)
        self.lineedit = widgets.LineEdit()
        self.lineedit.set_read_only()
        layout += self.lineedit
        action = widgets.Action()
        action.triggered.connect(self.choose_font)
        self.button = widgets.ToolButton()
        self.button.setDefaultAction(action)
        layout += self.button

    def __repr__(self):
        return f"FontChooserButton({self.current_font})"

    def __getstate__(self):
        return dict(font=self.current_font,
                    enabled=self.isEnabled())

    def __setstate__(self, state):
        self.__init__()
        if state["font"]:
            self.set_font(state["font"])
        self.set_enabled(state.get("enabled", True))

    @core.Slot()
    def choose_font(self):
        dlg = widgets.FontDialog()
        if self.current_font:
            dlg.setCurrentFont(self.current_font)

        if dlg.exec_():
            self.set_font(dlg.current_font())
            self.value_changed.emit(dlg.current_font())

    def set_font(self, font):
        if isinstance(font, str):
            self.current_font = gui.Font(font)
        else:
            self.current_font = font
        self.lineedit.setText(self.current_font.family())


if __name__ == "__main__":
    app = widgets.app()
    btn = FontChooserButton()
    btn.set_font("Consolas")
    btn.show()
    btn.value_changed.connect(print)
    app.exec_()



"""
Contains the quick reference widget
"""
from prettyqt import widgets
import pathlib


class QuickRefWidget(widgets.Widget):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.resize(608, 353)
        self.gridLayout = widgets.GridLayout(self)
        self.gridLayout.setObjectName("gridLayout")
        self.textedit_quickref = widgets.TextEdit(parent=self)
        self.textedit_quickref.setReadOnly(True)
        self.textedit_quickref.setObjectName("textedit_quickref")
        self.gridLayout.addWidget(self.textedit_quickref, 0, 0, 1, 1)
        html_file = pathlib.Path("ref.html")
        text = html_file.read_text()
        self.textedit_quickref.setHtml(text)

    def _show_context_menu(self, pos):
        self.context_menu.exec_(self.textedit_quickref.mapToGlobal(pos))


if __name__ == "__main__":
    app = widgets.app()
    widget = QuickRefWidget()
    widget.show()
    app.main_loop()



# -*- coding: utf-8 -*-
"""
"""

from qtpy import QtCore, QtWidgets

from prettyqt import core, widgets
from prettyqt.utils import bidict

ORIENTATIONS = bidict(horizontal=QtCore.Qt.Horizontal,
                      vertical=QtCore.Qt.Vertical)


QtWidgets.QScrollBar.__bases__ = (widgets.AbstractSlider,)


class ScrollBar(QtWidgets.QScrollBar):

    value_changed = core.Signal(int)

    def __init__(self, orientation="horizontal", parent=None):
        if orientation in ORIENTATIONS:
            orientation = ORIENTATIONS[orientation]
        super().__init__(orientation, parent)
        self.valueChanged.connect(self.on_value_change)



# -*- coding: utf-8 -*-
"""
"""

import logging

from prettyqt import core, widgets

logger = logging.getLogger(__name__)


class CheckBoxDelegate(widgets.ItemDelegate):
    """
    A delegate that places a fully functioning CheckBox in every
    cell of the column to which it's applied
    """

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.selected = []

    def createEditor(self, parent, option, index):
        """
        override
        instanciate the editor widget and initialize it
        also connect currentIndexChanged signal
        """
        cb = widgets.CheckBox(parent)
        cb.currentIndexChanged.connect(self.currentIndexChanged)
        return cb

    def setEditorData(self, cb, index):
        """
        override
        set correct initial value for editor widget
        """
        current_selection = index.data()
        with cb.block_signals():
            cb.setCurrentText(current_selection)

    def setModelData(self, combo, model, index):
        """
        override, gets called on self.commitData (?)
        apply the newly selected dtype to the column if possible
        """
        dtype = self.dtypes[combo.currentText()]
        # s = model.data(index, model.DATA_ROLE)
        try:
            model.setData(index, dtype, model.DTYPE_ROLE)
        except ValueError as e:
            logger.error(e)

    @core.Slot()
    def currentIndexChanged(self):
        self.commitData.emit(self.sender())



# -*- coding: utf-8 -*-
"""
"""

import functools
import logging
import operator
from typing import Any, Generator, List, Optional

from qtpy import QtCore, QtWidgets

from prettyqt import constants, gui, widgets
from prettyqt.utils import bidict

TRIGGERS = bidict(none=QtWidgets.QAbstractItemView.NoEditTriggers,
                  double_click=QtWidgets.QAbstractItemView.DoubleClicked,
                  edit_key=QtWidgets.QAbstractItemView.EditKeyPressed)

SELECTION_BEHAVIOURS = bidict(rows=QtWidgets.QAbstractItemView.SelectRows,
                              columns=QtWidgets.QAbstractItemView.SelectColumns,
                              items=QtWidgets.QAbstractItemView.SelectItems)

SELECTION_MODES = bidict(single=QtWidgets.QAbstractItemView.SingleSelection,
                         extended=QtWidgets.QAbstractItemView.ExtendedSelection,
                         multi=QtWidgets.QAbstractItemView.MultiSelection,
                         none=QtWidgets.QAbstractItemView.NoSelection)

SCROLL_MODES = bidict(item=QtWidgets.QAbstractItemView.ScrollPerItem,
                      pixel=QtWidgets.QAbstractItemView.ScrollPerPixel)

QtWidgets.QAbstractItemView.__bases__ = (widgets.AbstractScrollArea,)


class AbstractItemView(QtWidgets.QAbstractItemView):

    def __len__(self):
        if self.model() is not None:
            return self.model().rowCount()
        return 0

    def selectAll(self):
        """
        Override, we dont want to selectAll for too many items for performance reasons
        """
        if self.model() is None:
            return None
        if self.model().rowCount() * self.model().columnCount() > 1_000_000:
            logging.warning("Too many cells to select.")
            return None
        super().selectAll()

    def set_model(self, model: Optional[QtCore.QAbstractItemModel]):
        """
        delete old selection model explicitely, seems to help with memory usage
        """
        old_model = self.model()
        old_sel_model = self.selectionModel()
        if old_model is not None or model is not None:
            self.setModel(model)
        # if old_model:
        #     old_model.deleteLater()
        #     del old_model
        if old_sel_model:
            old_sel_model.deleteLater()
            del old_sel_model

    def set_delegate(self,
                     delegate: QtWidgets.QItemDelegate,
                     column: Optional[int] = None,
                     row: Optional[int] = None,
                     persistent: bool = False):
        if column is not None:
            self.setItemDelegateForColumn(column, delegate)
            if persistent:
                model = self.model()
                for i in range(0, model.rowCount()):
                    self.openPersistentEditor(model.index(i, column))
        elif row is not None:
            self.setItemDelegateForRow(row, delegate)
            if persistent:
                model = self.model()
                for i in range(0, model.columnCount()):
                    self.openPersistentEditor(model.index(row, i))
        else:
            self.setItemDelegate(delegate)

    def toggle_select_all(self):
        """
        select all items from list (deselect when all selected)
        """
        if self.selectionModel() is None:
            return None
        if self.selectionModel().hasSelection():
            self.clearSelection()
        else:
            self.selectAll()

    def set_table_color(self, color: str):
        self.setStyleSheet(f"QHeaderView::section {{ background-color:{color} }}")

    def current_index(self) -> Optional[int]:
        if self.selectionModel() is None:
            return None
        return self.selectionModel().currentIndex()

    def current_data(self):
        if self.model() is None:
            return None
        return self.current_index().data(QtCore.Qt.UserRole)

    def current_row(self) -> int:
        return self.current_index().row()

    def current_column(self) -> int:
        return self.current_index().column()

    def selected_indexes(self) -> List[QtCore.QModelIndex]:
        """
        returns list of selected indexes in first row
        """
        indexes = (x for x in self.selectedIndexes() if x.column() == 0)
        return sorted(indexes, key=lambda x: x.row())

    def selected_names(self) -> Generator[Any, None, None]:
        """
        returns generator yielding item names
        """
        return (x.data(constants.NAME_ROLE)
                for x in self.selected_indexes())

    def selected_rows(self) -> Generator[int, None, None]:
        """
        returns generator yielding row nums
        """
        return (x.row() for x in self.selected_indexes())

    def selected_data(self) -> Generator[Any, None, None]:
        """
        returns generator yielding selected userData
        """
        return (x.data(constants.USER_ROLE)
                for x in self.selected_indexes())

    def setup_dragdrop_move(self):
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDragDropMode(self.DragDrop)
        self.setDefaultDropAction(QtCore.Qt.MoveAction)
        self.setDropIndicatorShown(True)

    def set_edit_triggers(self, *triggers: Optional[str]):
        triggers = ["none" if t is None else t for t in triggers]
        for item in triggers:
            if item not in TRIGGERS:
                raise ValueError("trigger type not available")
        flags = functools.reduce(operator.ior, [TRIGGERS[t] for t in triggers])
        self.setEditTriggers(flags)

    def get_edit_triggers(self) -> list:
        return [k for k, v in TRIGGERS.items() if v & self.editTriggers()]

    def set_selection_behaviour(self, behaviour: str):
        """set selection behaviour for given item view

        Allowed values are "rows", "columns", "items"

        Args:
            behaviour: selection behaviour to use

        Raises:
            ValueError: behaviour does not exist
        """
        if behaviour not in SELECTION_BEHAVIOURS:
            raise ValueError("invalid selection behaviour")
        self.setSelectionBehavior(SELECTION_BEHAVIOURS[behaviour])

    def get_selection_behaviour(self) -> str:
        """returns current selection behaviour

        Possible values: "rows", "columns", "items"

        Returns:
            selection behaviour
        """
        return SELECTION_BEHAVIOURS.inv[self.selectionBehavior()]

    def set_selection_mode(self, mode: Optional[str]):
        """set selection mode for given item view

        Allowed values are "single", "extended", "multi" or "none"

        Args:
            mode: selection mode to use

        Raises:
            ValueError: mode does not exist
        """
        if mode is None:
            mode = "none"
        if mode not in SELECTION_MODES:
            raise ValueError("Format must be either 'single', 'extended',"
                             "'multi' or 'None'")
        self.setSelectionMode(SELECTION_MODES[mode])

    def get_selection_mode(self) -> str:
        """returns current selection mode

        Possible values: "single", "extended", "multi" or "none"

        Returns:
            selection mode
        """
        return SELECTION_MODES.inv[self.selectionMode()]

    def set_scroll_mode(self, mode: str):
        """sets the scroll mode for both directions

        possible values are "item", "pixel"

        Args:
            mode: mode to set

        Raises:
            ValueError: invalid scroll mode
        """
        if mode not in SCROLL_MODES:
            raise ValueError("Invalid scroll mode")
        self.setHorizontalScrollMode(SCROLL_MODES[mode])
        self.setVerticalScrollMode(SCROLL_MODES[mode])

    def set_horizontal_scroll_mode(self, mode: str):
        """sets the horizontal scroll mode

        possible values are "item", "pixel"

        Args:
            mode: mode to set

        Raises:
            ValueError: invalid scroll mode
        """
        if mode not in SCROLL_MODES:
            raise ValueError("Invalid scroll mode")
        self.setHorizontalScrollMode(SCROLL_MODES[mode])

    def set_vertical_scroll_mode(self, mode: str):
        """sets the vertical scroll mode

        possible values are "item", "pixel"

        Args:
            mode: mode to set

        Raises:
            ValueError: invalid scroll mode
        """
        if mode not in SCROLL_MODES:
            raise ValueError("Invalid scroll mode")
        self.setVerticalScrollMode(SCROLL_MODES[mode])

    def num_selected(self) -> int:
        """returns amount of selected rows

        Returns:
            amount of selected rows
        """
        if self.selectionModel() is None:
            return 0
        return len(self.selectionModel().selectedRows())

    def jump_to_column(self, col_num: int):
        """make sure column at given index is visible

        scrolls to column at given index

        Args:
            col_num: column to scroll to
        """
        if self.model() is None:
            return None
        idx = self.model().index(0, col_num)
        self.scrollTo(idx)

    def scroll_to_top(self):
        """override to use abstractitemview-way of scrolling to top
        """
        self.scrollToTop()

    def scroll_to_bottom(self):
        """override to use abstractitemview-way of scrolling to bottom
        """
        self.scrollToBottom()

    def select_last_row(self):
        idx = self.model().createIndex(self.model().rowCount() - 1, 0)
        self.setCurrentIndex(idx)

    def highlight_when_inactive(self):
        """also highlight items when widget does not have focus
        """
        p = gui.Palette()
        p.highlight_inactive()
        self.setPalette(p)
